#!/usr/bin/env python2
#
#  Valosoitin - Audio player for EFL
#
#  Copyright 2012-2013 Kai Huuhko <kai.huuhko@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import sys
import os
import time
import urllib
import argparse
import datetime
import cgi
import ConfigParser
import cPickle
import random
random.seed()
import mimetypes
mimetypes.init()

import mutagen

import elementary as elm
from ecore import timer_add, Idler
from emotion import Emotion

from ValoDBus import ValosoitinDBus
from gui import PlayControls, PlayModes, VolumeControls, PlayList, PlayListControls, PlayListItem, SeekControls, AudioInfo

parser = argparse.ArgumentParser(description='Utters sounds.')
parser.add_argument("files", metavar="file", type=str, nargs="*",
                    help="Files to add to the playlist")
clargs = parser.parse_args(sys.argv[1:])

import dbus, e_dbus
ml = e_dbus.DBusEcoreMainLoop()
dbus.set_default_main_loop(ml)
import dbus.service
bus = dbus.SessionBus()

dbo = None
try:
    dbo = bus.get_object("org.mpris.MediaPlayer2.Valosoitin", "/org/mpris/MediaPlayer2")
except dbus.exceptions.DBusException:
    pass

if dbo:
    if clargs.files:
        for f in clargs.files:
            dbo.OpenUri(f, dbus_interface="org.mpris.MediaPlayer2.Player")
    sys.exit()

conf = ConfigParser.SafeConfigParser()
conf_dir = os.path.expanduser(os.path.join("~", ".config", "valosoitin"))
conf_path = os.path.join(conf_dir, "valosoitin.conf")
if not os.path.exists(conf_dir):
    os.mkdir(conf_dir, 0700)
if not os.path.exists(os.path.join(conf_path)):
    conf.add_section("Settings")
    conf.set("Settings", "audio_backend", "gstreamer")
    conf.set("Settings", "single", "False")
    conf.set("Settings", "random", "False")
    conf.set("Settings", "repeat", "False")
    with open(conf_path, 'wb') as configfile:
        conf.write(configfile)
conf.read(conf_path)
try:
    if not conf.get("Settings", "audio_backend") in ("gstreamer", "xine"):
        print("Invalid audio backend in config file.")
        print("Set it to either gstreamer or xine.")
        sys.exit()
except:
    print("audio_backend was not found in settings file.")
    print("Set it to a value or remove the settings file so that")
    print("a new one will be generated at next start of this program.")

class Valosoitin(object):
    def __init__(self, *args):
        self.duration = datetime.timedelta()
        self.position = datetime.timedelta()
        self.seeking = False
        self.playtimer = None
        self.playlistitems = {}
        self.playlistindex = []
        self.randomorder = []

        window = self.window = elm.StandardWindow("valosoitin", "Valosoitin")
        window.callback_delete_request_add(self.quit)

        # GUI
        self.playcontrols = PlayControls(self)
        self.volumecontrols = VolumeControls(self)
        self.playmodes = PlayModes(self)
        self.seekcontrols = SeekControls(self)
        self.audioinfo = AudioInfo(self)
        self.playlist = PlayList(self)
        self.playlistcontrols = PlayListControls(self)

        # Top left box
        topleftbox = elm.Box(window)
        topleftbox.pack_end(self.playcontrols)
        topleftbox.pack_end(self.volumecontrols)
        topleftbox.pack_end(self.playmodes)
        topleftbox.show()

        # Top right box
        toprightbox = elm.Box(window)
        toprightbox.size_hint_weight_set(1.0, 0.0)
        toprightbox.size_hint_align_set(-1.0, 0.0)
        toprightbox.pack_end(self.seekcontrols)
        toprightbox.pack_end(self.audioinfo)
        toprightbox.show()

        # Top box
        topbox = elm.Box(window)
        topbox.horizontal_set(True)
        topbox.size_hint_weight_set(1.0, 0.0)
        topbox.size_hint_align_set(-1.0, 0.0)
        topbox.pack_end(topleftbox)
        topbox.pack_end(toprightbox)
        topbox.show()

        # Main box
        mainbox = elm.Box(window)
        mainbox.size_hint_weight_set(1.0, 1.0)
        window.resize_object_add(mainbox)
        mainbox.pack_end(topbox)
        mainbox.pack_end(self.playlist)
        mainbox.pack_end(self.playlistcontrols)
        mainbox.show()

        # Setup GUI update timer
        self.playtimer = playtimer = timer_add(0.2, self.update)
        playtimer.freeze()

        # Setup Emotion
        self.player = player = Emotion(window.evas_get(), module_filename=conf.get("Settings", "audio_backend"))
        player.video_mute_set(True)
        player.callback_add("playback_started", self.playing)
        player.callback_add("playback_finished", self.finished)
        self.volumecontrols.volume.value = player.audio_volume_get()*100
        self.volumecontrols.mute.state_set(player.audio_mute_get())

        self.dbusname = dbus.service.BusName("org.mpris.MediaPlayer2.Valosoitin", dbus.SessionBus())
        self.dbo = ValosoitinDBus(self)

        window.resize(400, 400)
        window.show()

        # Restore playlist
        try:
            pkl_file = open(os.path.expanduser(os.path.join("~", ".local", "share", "valosoitin.library")), 'rb')
        except IOError:
            pass
        else:
            try:
                paths = cPickle.load(pkl_file)
            except EOFError:
                pass
            else:
                path_generator = (f for f in paths)
                print(time.clock())
                idler = Idler(self.add_file, None, None, path_generator)
            pkl_file.close()

        # Add files from command line
        if clargs.files:
            path_generator = (f for f in clargs.files)
            print(time.clock())
            idler = Idler(self.add_file, None, None, path_generator)

    def raise_window(self):
        self.window.raise_()

    def update(self):
        pos = self.player.position_get()
        self.dbo.playerprops["Position"] = pos*1000000

        if not self.seeking:
            self.seekcontrols.seek.value = pos
        pos = datetime.timedelta(seconds=int(pos))
        self.position = pos
        t = "%s / %s" % (pos, self.duration)
        self.seekcontrols.seektime.text_set(t)

        dur = self.player.play_length
        if dur != self.duration:
            self.seekcontrols.seek.min_max_set(0, dur)
            dur = datetime.timedelta(seconds=int(dur))
            self.duration = dur

        return 1

    def add_file(self, fs, f, paths=None):
        if paths:
            try:
                f = paths.next()
            except StopIteration:
                self.randomize()
                print(time.clock())
                return False

        if not f:
            print("No file selected")
            return True

        if not os.path.exists(f):
            print("File doesn't exist")
            return True

        mimetype = mimetypes.guess_type(f, strict=False)[0]
        if mimetype not in (
            "application/ogg",
            "audio/basic",
            "audio/flac",
            "audio/mid",
            "audio/mp4",
            "audio/mpeg",
            "audio/ogg",
            "audio/vorbis",
            "audio/wav",
            "audio/x-wav",
        ):
            #print "invalid file", mimetype
            return True

        m = mutagen.File(f, easy=True)
        if not m:
            m = f

        # Playlist items are unique by filepath
        try:
            i = self.playlistindex.index(f)
        except ValueError:
            item = self.playlist.item_append(PlayListItem(), (self.player, m), None, elm.ELM_GENLIST_ITEM_NONE)
            self.playlistitems[f] = item
            self.playlistindex.append(f)
            # Recalculate random every time a file is added
            if not paths:
                self.randomize()
        else:
            n = elm.Notify(self.window)
            l = elm.Label(self.window)
            l.text = "Audio already on playlist"
            n.content = l
            l.show()
            n.show()
            n.timeout = 3.0

        return True

    def add_dir(self, fs, d):
        if not d:
            return
        paths = []
        for dirpath, dirnames, filenames in os.walk(d):
            for filename in filenames:
                paths.append(os.path.join(dirpath, filename))
        paths.sort()
        path_generator = (f for f in paths)
        print(time.clock())
        idler = Idler(self.add_file, fs, None, path_generator)

    def item_selected(self, genlist, item):
        player, m = item.data_get()
        if "mutagen" in str(type(m)):
            f = m.filename
        else:
            f = m
        i = self.playlistindex.index(f)+1
        l = len(self.playlistindex)
        self.playlistcontrols.plinfo.text_set("%s / %s" % (i, l))

    def play_selected(self, genlist, item):
        player, m = item.data_get()
        self.play_file(m)

    def play_file(self, m):
        if "mutagen" in str(type(m)):
            f = m.filename
        else:
            f = m

        current_file = self.player.file_get()
        current_item = None
        if current_file:
            try:
                current_item = self.playlistitems[urllib.unquote(current_file.split("://")[1])]
            except KeyError:
                pass

        new_item = self.playlistitems[f]

        self.player.file_set("file://"+urllib.quote(f))
        self.play()
        self.player.position = 0.0

        self.current_i = self.playlistindex.index(f)

        self.dbo.playerprops["Metadata"]["mpris:trackid"] = f

        if "mutagen" in str(type(m)):
            title = m.get("title")
            if not title:
                title = "Unknown Title"
            self.dbo.playerprops["Metadata"]["xesam:title"] = str(title)

            album = m.get("album")
            if not album:
                album = "Unknown Album"
            self.dbo.playerprops["Metadata"]["xesam:album"] = str(album)

            artist = m.get("artist")
            if not artist:
                artist = "Unknown Artist"
            self.dbo.playerprops["Metadata"]["xesam:artist"] = str(artist)

        #for v in self.playlist.realized_items_get():
            #v.fields_update("elm.swallow.icon", elm.ELM_GENLIST_ITEM_FIELD_CONTENT)

        if current_item:
            current_item.fields_update("elm.swallow.icon", elm.ELM_GENLIST_ITEM_FIELD_CONTENT)
        new_item.fields_update("elm.swallow.icon", elm.ELM_GENLIST_ITEM_FIELD_CONTENT)

    def play_pause(self, *args):
        if self.player.play_get():
            self.pause()
        else:
            self.play()

    def play(self):
        if self.player.play_get():
            return
        ic = elm.Icon(self.window)
        if not self.player.file_get():
            i = self.playlist.selected_item
            if not i:
                try:
                    f = self.playlistindex[0]
                except:
                    print("No files on the playlist")
                    return
                i = self.playlistitems[f]
            player, m = i.data_get()
            self.play_file(m)
            return
        self.player.play_set(True)
        ic.standard_set("media-playback-pause")
        self.playcontrols.play.content_set(ic)
        self.playcontrols.play.size_hint_min_set(30, 30)
        self.playtimer.thaw()
        self.dbo.playerprops["PlaybackStatus"] = "Playing"

    def pause(self, override=False):
        if not self.player.play_get() and not override:
            return
        ic = elm.Icon(self.window)
        self.player.play_set(False)
        ic.standard_set("media-playback-start")
        self.playcontrols.play.content_set(ic)
        self.playcontrols.play.size_hint_min_set(30, 30)
        self.playtimer.freeze()
        self.dbo.playerprops["PlaybackStatus"] = "Paused"

    def nx(self, btn, prev=False):
        f = self.player.file_get()
        if f:
            f = urllib.unquote(f).split("://")[1]
        else:
            return
        #current = self.playlistitems[f]
        c_i = self.current_i
        if self.playmodes.random.state:
            i = self.randomorder.index(c_i)
            if prev:
                i = i-1
                if i < 0:
                    if self.playmodes.repeat.state:
                        i = len(self.playlistindex)-1
                        #i = self.playlist.items_count-1
                    else:
                        return
            else:
                i = i+1
                if i > len(self.playlistindex)-1:
                #if i > self.playlist.items_count-1:
                    if self.playmodes.repeat.state:
                        i = 0
                    else:
                        return
            next_i = self.playlistindex[self.randomorder[i]]
            nextitem = self.playlistitems[next_i]
        else:
            if prev:
                next_i = c_i - 1
                if next_i >= 0:
                    nextitem = self.playlistitems[self.playlistindex[next_i]]
                else:
                    nextitem = None
            else:
                next_i = c_i + 1
                if next_i < len(self.playlistindex):
                    nextitem = self.playlistitems[self.playlistindex[next_i]]
                else:
                    nextitem = None
        if not nextitem:
            if self.playmodes.repeat.state:
                nextitem = self.playlistitems[self.playlistindex[0]]
            else:
                if not btn:
                    # End of playlist
                    self.pause(True)
                return
        self.play_selected(None, nextitem)

    def seekstart(self, slider):
        self.seeking = True

    def seektime(self, slider, t=None):
        self.seeking = False
        t = t if t is not None else slider.value
        ft = datetime.timedelta(seconds=int(t))
        d = self.duration
        self.seekcontrols.seektime.text_set("%s / %s" % (ft, d))
        self.player.position_set(t)
        self.dbo.Seeked(t*1000000)

    def seek_by(self, t):
        ct = self.player.position_get()
        t = ct+t
        self.seektime(None, t)

    def change_volume(self, slider):
        v = slider.value
        self.player.audio_volume_set(v/100)
        self.dbo.playerprops["Volume"] = v/100

    def mute(self, check):
        if check:
            v = check.state
            self.volumecontrols.volume.disabled_set(v)
            self.player.audio_mute_set(v)
        else:
            state = self.player.audio_mute_get()
            self.player.audio_mute_set(not state)
            self.volumecontrols.volume.disabled_set(not state)
            self.volumecontrols.mute.state = not state

    def set_play_mode(self, check):
        print(check.text)
        if check.text == "Single":
            if self.playmodes.repeat.state:
                if check.state:
                    v = "Track"
                else:
                    v = "Playlist"
            else:
                v = "None"
        elif check.text == "Repeat":
            if check.state:
                if self.playmodes.single.state:
                    v = "Track"
                else:
                    v = "Playlist"
            else:
                v = "None"
        print("setting to %s" % v)
        self.dbo.playerprops["LoopStatus"] = v

    def set_random_mode(self, check):
        self.dbo.playerprops["Shuffle"] = check.state

    def playing(self, player):
        current = self.playlistitems[urllib.unquote(self.player.file_get()).split("://")[1]]
        player, m = current.data_get()
        artists = ""
        albums = ""
        titles = ""
        if "mutagen" in str(type(m)):
            if m.get("artist"):
                artists = "by <i>"+cgi.escape(", ".join(m.get("artist")))+"</i>"
            if m.get("album"):
                albums = "from the album <b>"+cgi.escape(", ".join(m.get("album")))+"</b>"
            if m.get("title"):
                titles = "title <b>"+cgi.escape(", ".join(m.get("title")))+"</b>"
            self.audioinfo.artist.text_set(artists)
            self.audioinfo.album.text_set(albums)
        else:
            titles = str("title <b>%s</b>" % os.path.basename(m))
        self.audioinfo.title.text_set(titles)

    def finished(self, player):
        if self.playmodes.single.state:
            if self.playmodes.repeat.state:
                self.play()
            else:
                self.pause(True)
        else:
            self.nx(None)

    def randomize(self):
        # Create a list from 0 to length of playlist, then shuffle it
        o = [x for x in range(len(self.playlistitems))]
        random.shuffle(o)
        self.randomorder = o

    def quit(self, *args):
        output = open(os.path.expanduser(os.path.join("~", ".local", "share", "valosoitin.library")), 'wb')
        cPickle.dump(self.playlistindex, output)
        output.close()
        conf.set("Settings", "single", str(self.playmodes.single.state_get()))
        conf.set("Settings", "random", str(self.playmodes.random.state_get()))
        conf.set("Settings", "repeat", str(self.playmodes.repeat.state_get()))
        with open(conf_path, 'wb') as configfile:
            conf.write(configfile)
        elm.exit()

if __name__ == "__main__":
    elm.init()
    valosoitin = Valosoitin(None)
    elm.run()
    elm.shutdown()
